from django.core.exceptions import ObjectDoesNotExist
from django.db import transaction

from codelist.models import KonovaCode
from codelist.settings import CODELIST_BIOTOPES_ID, CODELIST_BIOTOPES_EXTRA_CODES_ID, \
    CODELIST_COMPENSATION_ACTION_DETAIL_ID, CODELIST_COMPENSATION_ACTION_ID, CODELIST_COMPENSATION_HANDLER_ID, \
    CODELIST_288_ID, CODELIST_AFTER_STATE_BIOTOPES_ID
from compensation.models import Compensation, CompensationState, CompensationAction, UnitChoices, CompensationDocument
from intervention.models import Responsibility, Intervention
from konova.management.commands.kspMigrater.base_migrater import BaseMigrater
from konova.models import Deadline, DeadlineType


class CompensationMigrater(BaseMigrater):

    def migrate(self):
        self.connect_db()
        cursor = self.db_connection.cursor()
        empty_str = "''"
        cursor.execute(
            'select '
                'om."KENNUNG", '
                'linf."OBJBEZ", '
                'ST_AsEWKT(ST_Multi(ST_CollectionExtract(ST_MakeValid(geomf.the_geom), 3))) as geomf, '
                'ST_AsEWKT(ST_Multi(ST_CollectionExtract(ST_MakeValid(geoml.the_geom), 2))) as geoml, '
                'ST_AsEWKT(ST_Multi(ST_CollectionExtract(ST_MakeValid(geomp.the_geom), 1))) as geomp, '
                'linf."Bemerkung" '
            'from "OBJ_MASTER" om '
                'left join "LINFOS" linf on om."GISPADID"=linf."GISPADID" '
                'left join kom on om."GISPADID"=kom.gispadid '
                'left join geometry_f geomf on om."GISPADID"=geomf.gispadid '
                'left join geometry_l geoml on om."GISPADID"=geoml.gispadid '
                'left join geometry_p geomp on om."GISPADID"=geomp.gispadid '
                'left join "Aufwertung" auf on om."GISPADID"=auf."GISPADID" '
            'where '
                'om."OKL"=7730080 and '
                'om.archiv=false and '
                f'(auf."Infos" is null or auf."Infos"={empty_str}) and '
                'om.nicht_vollstaendig=0 '
        )

        all_koms = cursor.fetchall()
        len_all_koms = len(all_koms)
        num_processed = 0
        print(f"Migrate KOMs to compensations...")
        print(f"--Found {len_all_koms} entries. Process now...")
        unsuccessfull_compensations = {}
        for kom in all_koms:
            if num_processed % 500 == 0:
                print(f"----{num_processed}/{len_all_koms} processed")

            with transaction.atomic():
                kom_identifier = kom[0]
                kom_title = kom[1]
                kom_comment = kom[5]
                compensation = Compensation.objects.get_or_create(
                    identifier=kom_identifier
                )[0]

                compensation.title = kom_title
                compensation.comment = kom_comment
                try:
                    compensation = self._migrate_interventions_reference(compensation, kom)
                except ObjectDoesNotExist:
                    compensation.delete()
                    unsuccessfull_compensations[kom_identifier] = "EIV does not exist"
                    continue

                compensation = self._migrate_par_7_data(compensation, kom)
                compensation = self._migrate_geometry(compensation, kom)
                compensation = self._migrate_responsibility(compensation, kom)
                compensation = self._migrate_compensation_type(compensation, kom)
                compensation = self._migrate_states(compensation, kom)
                compensation = self._migrate_deadlines(compensation, kom)
                compensation = self._migrate_action_control_deadlines(compensation, kom)
                compensation = self._migrate_actions(compensation, kom)
                compensation = self._migrate_log(compensation, kom)
                compensation = self._migrate_documents(compensation, CompensationDocument, kom)
                compensation.save()
                num_processed += 1
        print("The following KOMs could not be migrated: ")
        for kom, val in unsuccessfull_compensations.items():
            print(kom)
        cursor.close()

    def _migrate_interventions_reference(self, compensation, kom):
        kom_identifier = f"'{kom[0]}'"
        tmp_cursor = self.db_connection.cursor()
        tmp_cursor.execute(
            'select '
                'ref."REFERENZ" '
            'from "OBJ_MASTER" om '
            'left join "LINFOS" linf on om."GISPADID"=linf."GISPADID" '
            'left join "REFERENZ" ref on om."GISPADID"=ref."GISPADID" '
            'where '
                'om."OKL"=7730080 and '
                f'om."KENNUNG"={kom_identifier};'
        )
        eivs = tmp_cursor.fetchall()
        len_eivs = len(eivs)
        if len_eivs != 1:
            raise ObjectDoesNotExist(f"{kom_identifier} has {len_eivs} EIVs!")
        eiv = eivs[0]
        try:
            intervention = Intervention.objects.get(
                identifier=eiv[0]
            )
        except ObjectDoesNotExist:
            raise ObjectDoesNotExist(f"{kom_identifier} could not find migrated {eiv}")
        compensation.intervention = intervention
        tmp_cursor.close()
        return compensation

    def _migrate_responsibility(self, compensation, kom):
        compensation.responsible = compensation.responsible or Responsibility.objects.create()
        return compensation

    def _migrate_states(self, compensation, kom):
        kom_identifier = f"'{kom[0]}'"
        before_states_sql = 'select ' \
                'b."Biotoptyp", ' \
                'b."Flaeche", ' \
                'b."PKEY" ' \
            'from "BtypHtyp" b ' \
                'left join "OBJ_MASTER" om on om."GISPADID"=b."GISPADID" ' \
            'where ' \
                f'om."KENNUNG"={kom_identifier} and ' \
                'b."Biotoptyp"!=0 and '\
                'b."Biotoptyp" is not null '

        zero_str = "'0'"
        z_code_before_states_sql = 'select z."Zusatzcode" from "Zusatzcodes" z where z."FKEY"={}'
        after_states_sql = 'select ' \
                'z."Zielbiotoptyp", ' \
                'z."Flaeche", ' \
                'z."PKEY" ' \
            'from "Zielbiotoptypen" z ' \
                'left join "OBJ_MASTER" om on om."GISPADID"=z."GISPADID" ' \
            'where ' \
                f'om."KENNUNG"={kom_identifier} and '\
                f'z."Zielbiotoptyp"!={zero_str} and '\
                'z."Zielbiotoptyp" is not null '
        z_code_after_states_sql = 'select z."Z_Code" from "Z_Codes" z where z."FKEY"={}'

        compensation.before_states.all().delete()
        compensation = self._process_state_migration(compensation, before_states_sql, z_code_before_states_sql, compensation.before_states)
        compensation.after_states.all().delete()
        compensation = self._process_state_migration(compensation, after_states_sql, z_code_after_states_sql, compensation.after_states)
        return compensation

    def _process_state_migration(self, compensation, sql, z_code_sql, state_manager):
        tmp_cursor = self.db_connection.cursor()
        tmp_cursor.execute(sql)
        db_result = tmp_cursor.fetchall()

        for result in db_result:
            state_type = result[0]
            if state_type is None or len(str(state_type)) == 0:
                # garbage
                continue
            state_surface = result[1] or 0.0
            if state_surface == 0.0:
                # no surface, no real state -> garbage
                continue
            pkey_entry = f"'{result[2]}'"
            try:
                state_code = KonovaCode.objects.get(
                    atom_id=state_type,
                    code_lists__in=[CODELIST_BIOTOPES_ID]
                )
            except ObjectDoesNotExist:
                # Very old data might contain entries from this deprecated list
                try:
                    state_code = KonovaCode.objects.get(
                        atom_id=state_type,
                        code_lists__in=[CODELIST_AFTER_STATE_BIOTOPES_ID]
                    )
                except ObjectDoesNotExist:
                    raise ObjectDoesNotExist(f"{compensation.identifier} has unknown state code: {state_type}")
            tmp_cursor_z_code = self.db_connection.cursor()
            tmp_cursor_z_code.execute(z_code_sql.format(pkey_entry))
            z_code_results = tmp_cursor_z_code.fetchall()
            z_codes = []
            
            for z_code_result in z_code_results:
                z_code = KonovaCode.objects.filter(
                    atom_id=z_code_result[0],
                    code_lists__in=[
                        CODELIST_BIOTOPES_EXTRA_CODES_ID,
                        CODELIST_288_ID,
                    ],
                )
                z_codes += z_code
            tmp_cursor_z_code.close()

            try:
                _state_obj = state_manager.get(
                    biotope_type=state_code,
                    surface=state_surface
                )
            except ObjectDoesNotExist:
                _state_obj = CompensationState.objects.create(
                    biotope_type=state_code,
                    surface=state_surface,
                )
                state_manager.add(_state_obj)

            _state_obj.biotope_type_details.set(z_codes)

        tmp_cursor.close()
        return compensation

    def _migrate_compensation_type(self, compensation, kom):
        kom_identifier = f"'{kom[0]}'"
        tmp_cursor = self.db_connection.cursor()
        tmp_cursor.execute(
            'select '
                'mt.typ '
            'from "OBJ_MASTER" om '
                'left join massnahmetyp mt on om."GISPADID"=mt.gispadid '
            'where '
                f'om."KENNUNG"={kom_identifier}'
        )
        db_result = tmp_cursor.fetchall()
        if len(db_result) == 0:
            raise AssertionError(f"{kom_identifier} has no specification on compensation type (CEF, ...)")
        for comp_type in db_result:
            if comp_type == 705816:
                # regular compensation, do nothing
                pass
            elif comp_type == 705815:
                compensation.is_cef = True
            elif comp_type == 705817:
                compensation.is_coherence_keeping = True
            elif comp_type == 154156555:
                compensation.is_coherence_keeping = True
                compensation.is_cef = True
        tmp_cursor.close()
        return compensation

    def _migrate_action_control_deadlines(self, compensation, kom):
        kom_identifier = f"'{kom[0]}'"
        tmp_cursor = self.db_connection.cursor()
        tmp_cursor.execute(
            'select '
                'k."Datum", '
                'k."Typ", '
                'k."Ergebnis", '
                'k."Stelle", '
                'k."Bemerkung"  '
            'from "OBJ_MASTER" om '
                'left join "MASSN" m on m."GISPADID"=om."GISPADID" '
                'left join "Kontrolle" k on m."PKEY"=k."FKEY" '
            'where '
                f'om."KENNUNG"={kom_identifier}'
        )

        control_types = {
            707175: "Zielerreichungskontrolle",
            707176: "Zustandskontrolle",
            707174: "Ausführungskontrolle",
        }
        control_results = {
            153097: "nicht beurteilbar",
            153954: "vollständig umgesetzt",
            153956: "teilweise umgesetzt",
            153957: "eher nicht umgesetzt",
            153960: "mit leichten Mängeln",
            153963: "ohne Mängel",
            187136: "Ziel nicht erreicht",
            187137: "Ziel teilweise erreicht",
            187139: "Ziel vollständig umgesetzt",
        }

        db_results = tmp_cursor.fetchall()
        for result in db_results:
            control_date = result[0]
            if control_date is None:
                # useless data
                continue
            control_type = result[1]
            control_result = result[2]
            control_responsible = result[3]
            control_comment = result[4] or ""

            try:
                control_type = control_types[control_type]
            except KeyError:
                control_type = "Kein Eintrag"
            try:
                control_result = control_results[control_result]
            except KeyError:
                control_result = "Kein Eintrag"

            try:
                control_responsible = KonovaCode.objects.get(
                    atom_id=control_responsible,
                    is_selectable=True,
                    is_archived=False,
                    code_lists__in=[CODELIST_COMPENSATION_HANDLER_ID]
                )
                control_responsible = control_responsible.long_name
            except ObjectDoesNotExist:
                control_responsible = "Kein Eintrag"

            control_comment += f"\n\nKontrolltyp: {control_type}"
            control_comment += f"\nKontrollergebnis: {control_result}"
            control_comment += f"\nKontrollstelle: {control_responsible}"

            try:
                deadline = compensation.deadlines.get(
                    type=DeadlineType.CONTROL,
                    date=control_date,
                    comment=control_comment
                )
            except ObjectDoesNotExist:
                deadline = Deadline.objects.create(
                    type=DeadlineType.CONTROL,
                    date=control_date,
                    comment=control_comment
                )
                compensation.deadlines.add(deadline)

        tmp_cursor.close()
        return compensation

    def _migrate_deadlines(self, compensation, kom):
        compensation.deadlines.all().delete()
        kom_identifier = f"'{kom[0]}'"
        tmp_cursor = self.db_connection.cursor()
        tmp_cursor.execute(
            'select '
                'f.umsetzungbis, '
                'f.unterhaltungbis, '
                'f.artunterhaltungbis, '
                'f.artumsetzungbis '
            'from kom k '
                'left join fristen f on k.fristen=f.id '
                'left join "OBJ_MASTER" om on k.gispadid=om."GISPADID" '
            'where '
                f'om."KENNUNG"={kom_identifier}'
        )
        db_results = tmp_cursor.fetchall()

        for result in db_results:
            finish_until = result[0]
            maintain_until = result[1]
            maintain_until_type = result[2]
            finish_until_type = result[3]

            if finish_until_type == 941256187:
                finish_until_type = "Termin endgültig"
            elif finish_until_type == 889700393:
                finish_until_type = "Termin vorläufig"
            else:
                finish_until_type = "Kein Eintrag"

            if maintain_until_type == 807238388:
                maintain_until_type = "Terminart 'bis zum'"
            elif maintain_until_type == 23925195:
                maintain_until_type = "Terminart 'dauerhaft'"
            elif maintain_until_type == 322729640:
                maintain_until_type = "Terminart 'nicht erforderlich'"
            else:
                maintain_until_type = "Kein Eintrag zur Terminart"

            if finish_until is not None:
                try:
                    deadline = compensation.deadlines.get(
                        type=DeadlineType.FINISHED,
                        date=finish_until,
                        comment=finish_until_type
                    )
                except ObjectDoesNotExist:
                    deadline = Deadline.objects.create(
                        type=DeadlineType.FINISHED,
                        date=finish_until,
                        comment=finish_until_type
                    )
                    compensation.deadlines.add(deadline)

            if maintain_until is not None:
                try:
                    deadline = compensation.deadlines.get(
                        type=DeadlineType.MAINTAIN,
                        date=maintain_until,
                        comment=maintain_until_type
                    )
                except ObjectDoesNotExist:
                    deadline = Deadline.objects.create(
                        type=DeadlineType.MAINTAIN,
                        date=maintain_until,
                        comment=maintain_until_type
                    )
                    compensation.deadlines.add(deadline)

        tmp_cursor.close()
        return compensation

    def _migrate_actions(self, compensation, kom):
        compensation.actions.all().delete()

        kom_identifier = f"'{kom[0]}'"
        tmp_cursor = self.db_connection.cursor()
        tmp_cursor.execute(
            'select '
                'm."PKEY", '
                'm."Massn_RLP" as atom_id_parent, '
                'e."E_Massn" as atom_id_detail, '
                'm."Proz_Anteil" as menge, '
                'm."Massnahm_Typ" as einheit, '
                'm."MassnBeschreib" '
            'from "OBJ_MASTER" om '
                'left join "MASSN" m on om."GISPADID"=m."GISPADID" '
                'left join "Entwicklung" e on m."PKEY"=e."FKEY" '
            'where '
                f'om."KENNUNG"={kom_identifier} and '\
                'm."Massn_RLP"!=0 '
        )
        db_results = tmp_cursor.fetchall()

        for result in db_results:
            if result[0] is None:
                continue
            action_pkey = f"'{result[0]}'"
            selectable_action_parent = result[1]
            selectable_action = result[2]  # could be None!
            amount = result[3]
            unit = result[4]
            comment = result[5] or ""

            if selectable_action_parent in [999990, 888880, 777770, 0]:
                # is garbage
                selectable_action_parent = None
            if selectable_action in [999999, 888888, 777777, 0]:
                # is garbage
                selectable_action = None

            if selectable_action is None and selectable_action_parent is None:
                # a giant load of garbage. Step over
                continue

            if unit == 710113:
                unit = UnitChoices.m
            elif unit == 710119:
                # durchmesser - no such entries, can skip!
                pass
            elif unit == 710118:
                unit = UnitChoices.st
            elif unit == 710116:
                unit = UnitChoices.ha
            elif unit == 710114:
                unit = UnitChoices.m2
            elif unit == 710115:
                unit = UnitChoices.m3
            else:
                unit = UnitChoices.st
                comment += "\nDatenmigration: Kein Eintrag zur Mengeneinheit"

            if amount is None:
                amount = 0.0
                comment += "\nDatenmigration: Kein Eintrag zur Menge"

            tmp_cursor_z_code = self.db_connection.cursor()
            tmp_cursor_z_code.execute(
                'select '
                    'pe."Posi_E" as zusatzmerkmal '
                'from "Posi_E" pe '
                'where '
                    f'pe."FKEY"={action_pkey} and '
                    'pe."Posi_E"!=0 and '
                    'pe."Posi_E" is not null'
            )
            z_code_results = tmp_cursor_z_code.fetchall()
            z_codes = []
            for z_code_result in z_code_results:
                try:
                    z_code = KonovaCode.objects.get(
                        atom_id=z_code_result[0],
                        code_lists__in=[CODELIST_COMPENSATION_ACTION_DETAIL_ID],
                        is_selectable=True,
                        is_archived=False
                    )
                except ObjectDoesNotExist:
                    raise ObjectDoesNotExist(f"{z_code_result[0]}, {kom_identifier}")
                z_codes.append(z_code)

            action_code = selectable_action or selectable_action_parent
            try:
                action_code = KonovaCode.objects.get(
                    atom_id=action_code,
                    code_lists__in=[CODELIST_COMPENSATION_ACTION_ID],
                )
            except ObjectDoesNotExist:
                raise ObjectDoesNotExist(f"{action_code}, {kom_identifier}")
            try:
                action = compensation.actions.get(
                    action_type__in=[action_code],
                    amount=amount,
                    unit=unit,
                    comment=comment
                )
            except ObjectDoesNotExist:
                action = CompensationAction.objects.create(
                    amount=amount,
                    unit=unit,
                    comment=comment
                )
                compensation.actions.add(action)
            action.action_type.set([action_code])
            action.action_type_details.set(z_codes)

            tmp_cursor_z_code.close()


        tmp_cursor.close()
        return compensation

    def __migrate_par_7_action(self, compensation, kom):
        # Since this codelist is deprecated and rather small, we define it once for this migration as a simple map
        code_map = {
            265452837: "ökologische Verbesserung bestehender land- oder forstwirtschaftlicher Bodennutzung und landschaftlicher Strukturen",
            619127441: "Erhaltung und Verbesserung von Dauergrünland, insbesondere durch Beweidung",
            255952123: "Renaturierung von Gewässern",
            125425801: "Entsiegelung und Renaturierung von nicht mehr benötigten versiegelten Flächen im Innen- und Außenbereich",
            328264164: "Schaffung und Erhaltung größerer, zusammenhängender Biotopverbundstrukturen ",
            492185802: "Entwicklung und Wiederherstellung gesetzlich geschützter Biotope einschließlich des Verbunds zwischen einzelnen, benachbarten Biotopen",
            695727364: "Herstellung eines günstigen Erhaltungszustands eines Lebensraumtyps oder eines Vorkommens einer besonders geschützten Art",
            505765917: "Grunderwerb",
        }
        identifier = f"'{kom[0]}'"
        tmp_cursor = self.db_connection.cursor()
        tmp_cursor.execute(
            'select '
                'par7.massnahmenachpar7 '
            'from "OBJ_MASTER" om '
                'join massnahmennachpar7 par7 on om."GISPADID"=par7.gispadid '
            'where '
                f'om."KENNUNG"={identifier}'
        )
        fetch_results = tmp_cursor.fetchall()
        for result in fetch_results:
            code = result[0]
            try:
                code = code_map[code]
            except KeyError:
                code = f"Unbekannt ({code})"
            compensation.comment += f"\nMaßnahmen nach § 7 Abs. 3 LNatSchG: {code}"

        tmp_cursor.close()
        return compensation

    def __migrate_par_7_area_details(self, compensation, kom):
        # Since this codelist is deprecated and rather small, we define it once for this migration as a simple map
        code_map = {
            134510452: "Natura 2000-Gebiete",
            40451403: "Flächen für Maßnahmen zur Verbesserung des ökologischen Gewässerzustands nach Wasserrahmenrichtlinie ",
            726756838: "Flächen in geschützten Teilen von Natur und Landschaft",
            742579102: "Zu Kompensationszwecken vorgesehene Flächen in Landschafts- oder Grünordnungsplänen",
            239671518: "Bewirtschaftungs- und Pflegemaßnahmen zur dauerhaften Aufwertung des Naturhaushaltes und des Landschaftsbildes im Rahmen von produktionsintegrierten Maßnahmen (PIK) auf land- und forstwirtschaftlichen Flächen, auch außerhalb von Schutzgebieten",
            685985349: "ökologische Aufwertung von Waldbeständen für Waldrodungen",
        }
        identifier = f"'{kom[0]}'"
        tmp_cursor = self.db_connection.cursor()
        tmp_cursor.execute(
            'select '
                'g.gebietskulisse '
            'from "OBJ_MASTER" om '
                'join gebietskulisse g on om."GISPADID"=g.gispadid '
            'where '
                f'om."KENNUNG"={identifier}'
        )
        fetch_results = tmp_cursor.fetchall()
        for result in fetch_results:
            code = result[0]
            try:
                code = code_map[code]
            except KeyError:
                code = f"Unbekannt ({code})"
            compensation.comment += f"\nGebietskulisse nach § 7 Abs. 1 und 2 LNatSchG: {code}"

        tmp_cursor.close()
        return compensation

    def _migrate_par_7_data(self, compensation, kom):
        compensation = self.__migrate_par_7_action(compensation, kom)
        compensation = self.__migrate_par_7_area_details(compensation, kom)
        return compensation