from contextlib import contextmanager

from django.core.cache import cache
import time

LOCK_EXPIRE = 60 * 10  # Lock expires in 10 minutes


@contextmanager
def cache_lock(lock_id):
    """ Cache based lock e.g. for parallel celery processing

    Derived from https://docs.celeryq.dev/en/latest/tutorials/task-cookbook.html

    Use it like
    ...
    with cache_lock(lock_id, self.app.oid) as acquired:
        if acquired:
            do_mutexed_stuff()
    ...

    """
    timeout_at = time.monotonic() + LOCK_EXPIRE - 3
    # cache.add fails if the key already exists
    status = cache.add(lock_id, "LOCKED", LOCK_EXPIRE)
    try:
        yield status
    finally:
        # advantage of using add() for atomic locking
        if time.monotonic() < timeout_at and status:
            # don't release the lock if we exceeded the timeout
            # to lessen the chance of releasing an expired lock
            # owned by someone else
            # also don't release the lock if we didn't acquire it
            cache.delete(lock_id)
